/*
 *	The Remote User Information Daemon
 *
 *	(c) 1997--2010 Martin Mares <mj@ucw.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <sys/stat.h>
#include <time.h>
#include <utmp.h>
#include <signal.h>
#include <dirent.h>
#include <errno.h>

#include "nwho.h"

struct hostrec {
  struct hostrec *next;
  char name[32];
  time_t last_rec;			/* 0=down */
  u32 addr;
};

static int sock, port;
static struct hostrec *first_host;
static time_t now, last_local_scan;
static char hostname[64];

static void
net_init(char *name)
{
  struct hostent *h;
  struct sockaddr_in sa;

  if (name)
    {
      if (! (h = gethostbyname(name)))
	die("Failed to resolve %s", name);
    }
  else
    h = NULL;

  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sock < 0)
    die("Cannot create socket: %m");

  sa.sin_family = AF_INET;
  sa.sin_port = port = htons(NWHO_PORT);
  sa.sin_addr.s_addr = INADDR_ANY;
  if (bind(sock, (struct sockaddr *) &sa, sizeof(sa)) < 0)
    die("Cannot bind to UDP port %d: %m", NWHO_PORT);

  if (h)
    {
      memcpy(&sa.sin_addr.s_addr, h->h_addr, sizeof(sa.sin_addr.s_addr));
      if (connect(sock, (struct sockaddr *) &sa, sizeof(sa)) < 0)
	die("Cannot connect socket: %m");
    }

  if (gethostname(hostname, sizeof(hostname)) < 0)
    die("Unable to get my own host name: %m");
}

static void
scan_utmp(struct nwho_pkt *p, time_t now)
{
  struct utmp *u;
  int cnt = 0;
  struct stat st;
  char device[32];
  struct userinfo *h;
  char name[UT_NAMESIZE+10];
  time_t last;

  setutent();
  while ((u = getutent()) && cnt < MAX_USERS)
    if (u->ut_type == USER_PROCESS && u->ut_user[0])
      {
	h = &p->users[cnt];
	memcpy(name, u->ut_user, UT_NAMESIZE);
	name[9] = 0;
	strcpy(h->name, name);
	h->login_time = htonl(now - u->ut_time);
	snprintf(h->con, sizeof(h->con), "%s", u->ut_line);
	snprintf(device, sizeof(device), "/dev/%s", u->ut_line);
	if (stat(device, &st) < 0)
	  continue;
	h->mesg_y = !!(S_IWGRP & st.st_mode);
	last = st.st_atime;
	if (now < last)
	  last = now;
	h->idle_time = htonl(now - last);
	cnt++;
      }
  p->num_users = htonl(cnt);
}

static void
scan_load(struct nwho_pkt *p)
{
  int n, i, j[6];
  char buf[256];

  n = open("/proc/uptime", O_RDONLY);
  if (n < 0)
    return;
  if (read(n, buf, sizeof(buf)) <= 0)
    return;
  close(n);
  if (!sscanf(buf, "%d", &i))
    return;
  p->uptime = htonl(i);

  n = open("/proc/loadavg", O_RDONLY);
  if (n < 0)
    return;
  if (read(n, buf, sizeof(buf)) <= 0)
    return;
  close(n);
  if (sscanf(buf, "%d.%d%d.%d%d.%d", &j[0], &j[1], &j[2], &j[3], &j[4], &j[5]) != 6)
    return;
  p->avl[0] = htonl(j[0]*100 + j[1]);
  p->avl[1] = htonl(j[2]*100 + j[3]);
  p->avl[2] = htonl(j[4]*100 + j[5]);
}

static void
make_pkt(struct nwho_pkt *pkt)
{
  bzero(pkt, sizeof(*pkt));
  pkt->magic = htonl(NWHO_MAGIC);
  pkt->local_time = htonl(now);
  scan_utmp(pkt, now);
  scan_load(pkt);
}

static void
cleanup(void)
{
  DIR *d;
  struct dirent *e;

  d = opendir(".");
  if (!d)
    die("opendir: %m");
  while (e = readdir(d))
    if (e->d_name[0] != '.')
      unlink(e->d_name);
  closedir(d);
}

static void
save_pkt(char *name, struct nwho_pkt *pkt)
{
  int len = nwho_pkt_size(pkt);
  int fd = open(name, O_WRONLY | O_CREAT, 0666);
  if (fd < 0)
    {
      syslog(LOG_ERR, "open(%s): %m", name);
      return;
    }
  pkt->server_time = htonl(now);
  if (write(fd, pkt, len) != len)
    syslog(LOG_ERR, "write: %m");
  ftruncate(fd, len);
  close(fd);
}

static inline int			/* Validation checks not implemented yet */
is_valid(u32 ipa)
{
  return 1;
}

static void
receive(void)
{
  int r;
  struct nwho_pkt pkt;
  struct sockaddr_in sa;
  socklen_t al = sizeof(sa);
  int n = sizeof(struct nwho_pkt) - MAX_USERS * sizeof(struct userinfo);
  struct hostrec *e;
  struct hostent *h;
  char *c;

  alarm(DEFAULT_PRUNE_TIME);
  r = recvfrom(sock, &pkt, sizeof(pkt), 0, (struct sockaddr *) &sa, &al);
  alarm(0);
  now = time(NULL);
  if (r < 0)
    {
      if (errno == EINTR)
	return;
      syslog(LOG_ERR, "recvfrom: %m");
      sleep(5);
      return;
    }

  if (!is_valid(sa.sin_addr.s_addr) || sa.sin_port != port)
    {
      syslog(LOG_WARNING, "Received packet from invalid source %s:%d", inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
      return;
    }

  if (r < n ||
      pkt.magic != htonl(NWHO_MAGIC) ||
      ntohl(pkt.num_users) > MAX_USERS ||
      r < nwho_pkt_size(&pkt))
    {
      syslog(LOG_WARNING, "Malformed packet from %s", inet_ntoa(sa.sin_addr));
      return;
    }

  for(e=first_host; e; e=e->next)
    if (e->addr == sa.sin_addr.s_addr)
      break;

  if (!e)
    {
      e = malloc(sizeof(struct hostrec));
      if (!e)
	{
	  syslog(LOG_ERR, "Out of memory");
	  return;
	}
      e->next = first_host;
      first_host = e;
      h = gethostbyaddr((char *) &sa.sin_addr, sizeof(sa.sin_addr), AF_INET);
      if (h)
	{
	  snprintf(e->name, sizeof(e->name), "%s", h->h_name);
	  for (c=e->name; *c; c++)
	    if (*c == '.')
	      {
		*c = 0;
		break;
	      }
	    else if ((*c < 'A' || *c > 'Z') /* Filter out malicious characters */
		     && (*c < 'a' || *c > 'z')
		     && (*c < '0' || *c > '9')
		     && *c != '-')
	      *c = '_';
	}
      else
	strcpy(e->name, inet_ntoa(sa.sin_addr));
      e->addr = sa.sin_addr.s_addr;
    }

  e->last_rec = now;
  save_pkt(e->name, &pkt);
}

static void
local_scan(void)
{
  struct nwho_pkt pkt;

  make_pkt(&pkt);
  save_pkt(hostname, &pkt);
}

static void
do_tick(void)
{
  struct hostrec *e;

  if (last_local_scan + DEFAULT_SEND_TIME <= now)
    {
      last_local_scan = now;
      local_scan();
    }
  for(e=first_host; e; e=e->next)
    if (e->last_rec && e->last_rec + DEFAULT_DEAD_TIME < now)
      {
	if (unlink(e->name) < 0)
	  syslog(LOG_ERR, "unlink(%s): %m", e->name);
	e->last_rec = 0;
      }
}

static void
daemonize(void)
{
  pid_t pid = fork();
  if (pid < 0)
    die("Fork failed: %m");
  if (pid)
    exit(0);

  close(0);
  close(1);
  close(2);
  open("/dev/null", O_RDONLY);
  dup(0);
  dup(0);
  setsid();
}

static void
tick(int unused)
{
}

static void
server(void)
{
  static struct sigaction sigact;

  net_init(NULL);
  if (chdir(NWHO_SPOOL_DIR) < 0)
    die("chdir(" NWHO_SPOOL_DIR "): %m");
  cleanup();
  daemonize();
  bzero(&sigact, sizeof(sigact));
  sigact.sa_handler = tick;
  sigaction(SIGALRM, &sigact, NULL);
  now = time(NULL);
  for(;;)
    {
      do_tick();
      receive();
    }
}

static void
client(char *serv)
{
  struct nwho_pkt pkt;

  net_init(serv);
  utmpname(UTMP_FILE);
  daemonize();
  for(;;)
    {
      now = time(NULL);
      make_pkt(&pkt);
      if (send(sock, &pkt, nwho_pkt_size(&pkt), 0) < 0)
	syslog(LOG_ERR, "sendmsg: %m");
      sleep(DEFAULT_SEND_TIME);
    }
}

int
main(int argc, char **argv)
{
  if (argc == 2 && !strcmp(argv[1], "--version"))
    {
      printf("nwho " STRINGIFY(VERSION) "\n");
      return 0;
    }

  if (argc == 2)
    client(argv[1]);
  else if (argc == 1)
    server();
  else
    {
      fprintf(stderr, "Usage: nwhod [<server>]\n");
      return 1;
    }
  return 0;
}
